// Paul Houle (paulhoule.com) March 22, 2008
//
//  This file is part of par2cmdline (a PAR 2.0 compatible file verification and
//  repair tool). See http://parchive.sourceforge.net for details of PAR 2.0.
//
//  par2cmdline is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  par2cmdline is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include "par2cmdline.h"
#include "par2x86.h"				// $$20171110: move x86 code to .asm file

/*----------------------------------------------------------------------------
;
; Does nothing if PrCap is non-zero; otherwise:
;	sets PrCap to processor capabilities we care about.
;	sets TotThr to number of total threads to use.
; Only needs to be called once immediately after parsing command line.
;
;---------------------------------------------------------------------------*/

int PrCap = 0;
static int TotThr = 0;

void SetPr(const CommandLine &commandline) {

	if (PrCap) return;
	PrCap = Asm_getpr();				// get processor capabilities

	/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	; $$20171110: By default 1 thread per processor core is created.
	; This may be overridden via command line option.
	*/

	if (PrCap & GETPR_MMX) {			// if MMX, allow mult processors
		int i = commandline.GetThreadCount(); // forced thread cnt
		if (i == 0) {
			SYSTEM_INFO si;
			GetSystemInfo(&si);			// ok for Win95+, all WinNT
			i = si.dwNumberOfProcessors; // fetch # of processors
		}
		TotThr = i;
	} else {							// else non-MMX,
		if (commandline.GetNoiseLevel() > CommandLine::nlQuiet) {
			cout << "*** Warning: non-MMX processor ***" << endl;
		}
	}
}

/*----------------------------------------------------------------------------
;
; Threading logic (init, release control struct) and thread function itself.
;
;---------------------------------------------------------------------------*/

// $$20171110 (moved to .h)	 #define MAXTH 32			// maximum threads

typedef struct {						// RS threading control struct
	unsigned init;						// non-zero after struct initialized

	u32 datapresent;					//	invariant (per inputindex) fields
	u32 datamissing;					//	.			.			.
	u32 inputindex;						//	.			.			.
	Galois16 *leftmatrix;				//	.			.			.
	Galois16 *table;					//	.			.			.
	char *inputbuffer;					//	.			.			.
	char *outputbuffer;					//	.			.			.
	size_t blocklength;					//	.			.			.
	size_t chunksize;					//	.			.			.
	u32 totalblockcount;				//	.			.			.

	unsigned nThreads;					// created threads

	unsigned blksrepdone;				// blocks reported done
				// volatile fields (may be shared among multiple threads)
	long volatile bltstart;				// blocks left to start
	long volatile bltfinish;			// blocks left to finish

				// kernel objects
	HANDLE sem;							// semaphore used to dispatch threads
	HANDLE ev;							// event used to sense completion
	HANDLE tH[MAXTH];					// vector of thread handles
} RSTH;
RSTH Th = {0};

// This routine processes 1 combined multiplication table and the block it
// applies to.  This is the atomic unit of work done by a thread.
//
// The *pTh structure must be initialized with the invariant fields (only
// the outputindex varies across the default 100 blocks).

void DoBlock(RSTH * const pTh, u32 outputindex
) {
	char Mtab_space[256*2*4+15];		// mult tables (+ extra for alignment)
	unsigned *pMtab;					// ptr to true start of Mtab
	size_t bsize;						// number of bytes to process
	char *inbuf;						// working input  block ptr
	char *outbuf;						// working output block ptr

	// Look up the appropriate element in the RS matrix

	Galois16 factor = pTh->leftmatrix[
	  outputindex * (pTh->datapresent + pTh->datamissing) + pTh->inputindex];
	// Do nothing if the factor happens to be 0
	if (factor == 0) return;

	pMtab = (unsigned *)				// para align true Mtab start addr
		(((size_t) Mtab_space + 15) & ~15);

	Asm_BuildMtab(						// build Mtab (combined mult tables)
		pMtab, factor, pTh->table);

	inbuf = pTh->inputbuffer;			// init in/out addr's
	outbuf = pTh->outputbuffer + (pTh->chunksize * outputindex);
	bsize = pTh->blocklength;			// init bytes to process

	for (;;) {

		// Handle tiny blocks with straightforward 4-byte routine -- for such
		// blocks it's faster to avoid the 8-byte routine's setup overhead.
		//
		// (the "40" number currently below was determined by time testing).

		if (bsize <= 40) {				// if too small to get fancy with,
			if (bsize) {				// if something at all,
				Asm_DoBlock4(inbuf, outbuf, bsize, pMtab);
			}
			break;
		}

		// inbuf & outbuf are always relatively 8-byte aligned, but the caller
		// operates in dword sizes so they may not be absolutely 8-byte
		// aligned.  Misalignment slows things down quite a bit (10%); so, if
		// addresses misaligned, process a dword beforehand to align them.

		if ((size_t) outbuf & 4) {		// if out misaligned (in will be too),
			Asm_DoBlock4(inbuf, outbuf, 4, pMtab);
			inbuf += 4;
			outbuf += 4;
			bsize -= 4;
		}

		// Process main portion of the data.  Details: (1) the Asm_DoBlock8
		// code reads ahead 8-bytes, so we limit it to not read beyond the
		// inputbuffer; (2) there may be a residual dword at the end.

		{
			size_t bsz = bsize & ~4;	// size must be a multiple of 8
			bsz -= 8;					// compensate for read-ahead
			Asm_DoBlock8(				// process large chunk 8-bytes a shot
				inbuf, outbuf, bsz, pMtab);
			Asm_DoBlock4(				// process residual 8-12 bytes
				inbuf + bsz, outbuf + bsz, bsize - bsz, pMtab);
		}
		break;
	}
}

void RelRSthrd(RSTH * const pTh) {	// release RS threading resources
	/*$$20150928 int */ unsigned i;
	
	if (pTh->nThreads) {				// if thread(s) were created,
		pTh->bltstart = -999999;		// command all threads to exit
		ReleaseSemaphore(pTh->sem, pTh->nThreads, NULL);
		WaitForMultipleObjects(			// wait for all threads to exit
			pTh->nThreads, pTh->tH, TRUE, 5000);
		for (i = 0; i < pTh->nThreads; i++) {
			CloseHandle(pTh->tH[i]);	// release thread handles
		}
	}
	if (pTh->ev) CloseHandle(pTh->ev);	// close event & semaphore handles
	if (pTh->sem) CloseHandle(pTh->sem);
}

// Note: the main thread function (below) and the work function it calls
// do not use any C run-time functions; hence, we can use CreateThread
// instead of beginthreadex.

DWORD WINAPI ThreadFunc(LPVOID targ) {	// top-level thread function
	RSTH * const pTh = (RSTH * const) targ;
	long bleft;

	do {
		WaitForSingleObject(pTh->sem, INFINITE); // await triggering
		do {
			bleft =						// number of blocks left to start
				InterlockedDecrement(&pTh->bltstart);
			if (bleft < 0) break;

			DoBlock(pTh,				// process the block
				pTh->totalblockcount - (bleft + 1));
			bleft =						// number of blocks left to finish
				InterlockedDecrement(&pTh->bltfinish);
		} while (bleft);
		if (bleft == 0) {				// if finished last block of set,
			SetEvent(pTh->ev);			// need to wake up the main thread
		}
	} while (bleft > -999999);			// operate till commanded to exit

	return 0;
}

void DoRelRSthrd(void) {				// needed for atexit() in InitRSthrd
	RelRSthrd(&Th);
}

// Create one thread per available extra core.
// If multi-processor operation is not possible sets pTh.nThreads == 0.

void InitRSthrd(RSTH * const pTh) {		// init threading resources
	int i, j;
	long volatile iitest = 99;			// for InterlockedIncrement test

	i = TotThr;							// target count of total threads
	--i;								// -1 since primary thread counts too
	for (;;) {
		if (i <= 0) break;				// need at least two processors
		// need an OS w/modern InterlockedIncrement (> Win 95 & NT 3.51)
		if (InterlockedIncrement(&iitest) != 100) break;

		i = (i <= MAXTH) ? i : MAXTH;	// limit to our max processors
		pTh->sem = CreateSemaphore(		// semaphore to dispatch threads
			NULL, 0, i * 2, NULL);		//	(non-signaled)
		if (!pTh->sem) break;
		pTh->ev = CreateEvent(			// event to sense end of processing
			NULL, FALSE, FALSE, NULL);	//	(non-signaled, auto-reset)
		if (!pTh->ev) break;
		for (j = 0; j < i; j++) {		// create threads for extra processors
			DWORD ThreadID;
			pTh->tH[j] = CreateThread(
				NULL, 2048, ThreadFunc, pTh, 0, &ThreadID);
			if (!pTh->tH[j]) break;
			pTh->nThreads++;
		}
		break;
	}
	if (iitest == 100) {				// if we tried to create threads,
		if (i == pTh->nThreads) {		//	and were entirely successful,
			atexit(DoRelRSthrd);		// install at-exit cleanup routine
		} else {						// else we tried and failed,
			RelRSthrd(pTh);				// kill anything halfway done
			pTh->nThreads = 0;			// disable threading
		}
	}
}

/*----------------------------------------------------------------------------
;
; Paul Houle (paulhoule.com) March 22, 2008
; The below routine is the gist of the changes.
; It runs the assembly code to handle the main CPU work of par2 creation,
; dispatching multiple threads (if there are cores to do so).
;
; This routine may not be called again (with outputindex != 0) after it has
; reported all blocks in a set have been processed.
;
;---------------------------------------------------------------------------*/

u32 ReedSolomon<Galois16>::AsmProcess(size_t size, u32 inputindex,
 const void *inputbuffer, u32 outputindex, void *outputbuffer,
 size_t chunksize, u32 totalblockcount
) {
	long bleft;

	// First call of many (typically 100) operating within same buffer?

	if (outputindex == 0) {

		Th.datapresent = datapresent;	// copy invariant fields needed by
		Th.datamissing = datamissing;	//	threads (threads have no access
		Th.leftmatrix = leftmatrix;		//	to the ReedSolomon instance)
		Th.table = glmt->tables;
		Th.blocklength = size;
		Th.inputbuffer = (char *) inputbuffer;
		Th.inputindex = inputindex;
		Th.outputbuffer = (char *) outputbuffer;
		Th.chunksize = chunksize;
		Th.totalblockcount = totalblockcount;

		// Set number of blocks that must be started and finished.

		Th.bltstart = totalblockcount;
		Th.bltfinish = totalblockcount;

		Th.blksrepdone = 0;				// count blocks reported done

		if (!Th.init) {					// if haven't created threads,
			Th.init++;					// only run the init code once
			InitRSthrd(&Th);			// create 1 thread per extra core
		}

		if (Th.nThreads					// if we have extra threads,
		 && totalblockcount > 1) {		//	and block(s) to give them,
			u32 ttr = totalblockcount - 1;
			if (ttr > Th.nThreads) ttr = Th.nThreads;
			ReleaseSemaphore(			// release thread(s) to process blocks
				Th.sem, ttr, NULL);
		}
	}

	bleft =								// number of blocks left to start
		InterlockedDecrement(&Th.bltstart);
	if (bleft < 0) {					// if nothing left to do,
		WaitForSingleObject(			// wait for last block(s) to finish
			Th.ev, INFINITE);
		bleft = 0;						// indicate all blocks are done
	} else {
		DoBlock(						// got a block to do, process it
			&Th, Th.totalblockcount - (bleft + 1));
		bleft =							// number of blocks left to finish
			InterlockedDecrement(&Th.bltfinish);
	}

	bleft = Th.totalblockcount - bleft;	// total blocks completed
	bleft -= Th.blksrepdone;			// blocks sensed completed this call
	Th.blksrepdone += bleft;			// update total

	Asm_emms();							// execute emms opcode
	return (u32) bleft;
}
